# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Models for the LiquidNewton contract for a given validator.  Note
that this contract also exposes functionality for claiming fees.
"""

from autonity.abi_manager import ABIManager
from autonity.erc20 import ERC20

from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.types import Wei, ChecksumAddress


class LiquidNewton(ERC20):
    """
    The LiquidNewton contract.

    This is intended for internal use only. The naming mirrors
    the LiquidNewton contract in the Autonity code, but it's not very
    intuitive to have methods such as "claim_rewards" on a Token.
    Rather, this is more logically exposed by the Validator class.

    TODO: review this API (here and in the Autonity contracts).
    """

    def __init__(self, web3: Web3, address: ChecksumAddress):
        super().__init__(web3, address, ABIManager.load_abi("Liquid"))

    def unclaimed_rewards(self, account: ChecksumAddress) -> Wei:
        """
        See function `unclaimedRewards` on LiquidNewton.sol
        """
        return self.contract.functions.unclaimedRewards(account).call()

    def claim_rewards(self) -> ContractFunction:
        """
        Create a ContractFunction to claim rewards.  Use
        `create_contract_function_transaction` to use this in a
        transaction.  See function `claimRewards` on LiquidNewton.sol.
        """
        return self.contract.functions.claimRewards()



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Functions for working with contract ABIs.
"""


from __future__ import annotations
from web3 import Web3
from web3.types import (
    ABI,
    ABIFunction,
    ABIFunctionParams,
)
from typing import Dict, List, Tuple, Sequence, Any, Union, Callable, cast


def find_abi_constructor(abi: ABI) -> ABIFunction:
    """
    Given an ABI and function name, find the ABIFunction element.
    """

    for element in abi:
        if element["type"] == "constructor" and "name" not in element:
            return cast(ABIFunction, element)

    raise ValueError("constructor not found in ABI")


def find_abi_function(abi: ABI, function_name: str) -> ABIFunction:
    """
    Given an ABI and function name, find the ABIFunction element.
    """

    for element in abi:
        if element["type"] == "function" and element["name"] == function_name:
            return cast(ABIFunction, element)

    raise ValueError(f"function {function_name} not found in ABI")


def parse_arguments(abi_function: ABIFunction, arguments: List[str]) -> List[Any]:
    """
    Given an ABI, a function name and a list of string parameters,
    parse the parameters to suitable types to call the function on a
    contract with the given ABI.
    """

    inputs = abi_function["inputs"]
    parsers = _argument_parsers_for_params(inputs)
    if len(parsers) != len(arguments):
        raise ValueError(f"function requires {len(parsers)}, received {len(arguments)}")

    return [parse(arg) for parse, arg in zip(parsers, arguments)]


def parse_return_value(abi_function: ABIFunction, return_value: Any) -> Any:
    """
    Top level entry point, given the ABI outputs.  Supports void
    types, and flattening a one-element tuple to a raw type.
    """

    outputs = abi_function["outputs"]
    if len(outputs) == 0:
        return None

    if len(outputs) == 1:
        # Single return value (including an array)
        return _parse_return_value_from_type(
            outputs[0]["type"], outputs[0], return_value
        )

    assert isinstance(return_value, tuple)
    return _parse_return_value_tuple(outputs, return_value)


ParamType = Union[str, int, float, bool]
"""
A native type which can be passed as an argument to a ContractFunction
"""

ParamParser = Callable[[str], ParamType]
"""
Function to parse a string to a ParamType
"""


def _parse_string(value: str) -> str:
    """
    Identity parser.
    """
    return value


def _parse_bool(bool_str: str) -> bool:
    """
    Boolean parser.
    """
    if bool_str in ["False", "false", "0", ""]:
        return False

    return True


def _string_to_argument_fn_for_type(arg_type: str) -> ParamParser:
    """
    Return a function which parses a string into a type suitable for
    function arguments.
    """

    # TODO: support complex types

    if "[" in arg_type:
        raise ValueError(f"cannot convert array type '{arg_type}' from string")

    if arg_type.startswith("uint") or arg_type.startswith("int"):
        return int
    if arg_type == "bool":
        return _parse_bool
    if arg_type == "address":
        return Web3.to_checksum_address
    if arg_type.startswith("bytes") or arg_type == "string":
        return _parse_string
    if arg_type.startswith("fixed") or arg_type.startswith("ufixed"):
        return float

    raise ValueError(f"cannot convert '{arg_type}' from string")


def _argument_parsers_for_params(
    outputs: Sequence[ABIFunctionParams],
) -> List[ParamParser]:
    """
    Given the ABIFunctionParams object representing the output types
    of a specific function, return a list of string-to-paramtype
    converters.
    """

    out_types: List[ParamParser] = []
    for output in outputs:
        out_types.append(_string_to_argument_fn_for_type(output["type"]))

    return out_types


def _parse_return_value_from_type(
    type_name: str, output: ABIFunctionParams, value: Any
) -> Any:
    """
    Parse a single value from an ABIFunctionParams.
    """

    # Check for array types
    if type_name.endswith("[]"):
        assert isinstance(value, list)
        element_type = type_name[:-2]
        return [_parse_return_value_from_type(element_type, output, v) for v in value]

    # Check for tuples
    if type_name == "tuple":
        assert isinstance(value, tuple)
        assert "components" in output
        return _parse_return_value_tuple(output["components"], value)

    return value


def _parse_return_value_as_anonymous_tuple(
    outputs: Sequence[ABIFunctionParams], values: Tuple
) -> Tuple:
    """
    Parse a list of unnamed ABIFunctionParams and a tuple, to a tuple.
    """
    assert len(values) == len(outputs)
    return tuple(
        _parse_return_value_from_type(out["type"], out, val)
        for val, out in zip(values, outputs)
    )


def _parse_return_value_as_named_tuple(
    outputs: Sequence[ABIFunctionParams], values: Tuple
) -> Dict[str, Any]:
    """
    Parse a list of named ABIFunctionParams and a tuple, to a dict.
    """

    assert len(values) == len(outputs)
    value_dict: Dict[str, Any] = {}
    for val, out in zip(values, outputs):
        value_dict[out["name"]] = _parse_return_value_from_type(out["type"], out, val)

    return value_dict


def _parse_return_value_tuple(
    outputs: Sequence[ABIFunctionParams], values: Tuple
) -> Any:
    """
    Anonymous tuples to tuples, named tuples to dictionaries.
    """

    assert len(values) == len(outputs)
    if outputs[0]["name"] == "":
        return _parse_return_value_as_anonymous_tuple(outputs, values)

    return _parse_return_value_as_named_tuple(outputs, values)



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Python module containing the Tendermint Web3 external module.
"""

from __future__ import annotations

from autonity.committee_member import CommitteeMember

import json
from web3.types import ABI, RPCEndpoint
from web3.module import Module
from web3.method import Method
from typing import Sequence, Tuple, Any, cast


class Tendermint(Module):
    """
    A custom Web3.module.Module that enables a JSON-RPC method call to
    `tendermint_getContractABI`.
    """

    def _getCommittee_munger(  # pylint: disable=invalid-name
        self, block_height: int
    ) -> Tuple[str]:
        return (("0x" + str(block_height)),)

    def _getCommitteeAtHash_munger(  # pylint: disable=invalid-name
        self, block_hash: bytes
    ) -> Tuple[str]:
        return (self.w3.to_hex(block_hash),)

    _getCommittee: Method = Method(
        json_rpc_method=RPCEndpoint("tendermint_getCommittee"),
        mungers=[_getCommittee_munger],
    )

    _getCommitteeAtHash: Method = Method(
        json_rpc_method=RPCEndpoint("tendermint_getCommitteeAtHash"),
        mungers=[_getCommitteeAtHash_munger],
    )

    _getContractABI: Method = Method(
        json_rpc_method=RPCEndpoint("tendermint_getContractABI")
    )

    _getContractAddress: Method = Method(
        json_rpc_method=RPCEndpoint("tendermint_getContractAddress")
    )

    _getCoreState: Method = Method(
        json_rpc_method=RPCEndpoint("tendermint_getCoreState")
    )

    _getCommitteeEnodes: Method = Method(
        json_rpc_method=RPCEndpoint("tendermint_getCommitteeEnodes")
    )

    def get_committee(self, block_height: int) -> Sequence[CommitteeMember]:
        """
        Return the committee at the given block height.
        """
        result = self._getCommittee(block_height)
        assert isinstance(result, list)
        return result

    def get_committee_at_hash(self, block_hash: bytes) -> Sequence[CommitteeMember]:
        """
        Return the committee at the given block hash.
        """
        result = self._getCommitteeAtHash(block_hash)
        print(f"result={result}")
        assert isinstance(result, list)
        committee_members = cast(Sequence[CommitteeMember], result)
        assert isinstance(result[0].address, str)
        assert isinstance(result[0].votingPower, int)
        return committee_members

    def get_contract_abi(self) -> ABI:
        """
        Returns the ABI of the Autonity contract.
        """
        abi_str = self._getContractABI()
        abi = json.loads(abi_str)
        assert isinstance(abi, list), f"ABI type {type(abi)})"
        return abi

    def get_contract_address(self) -> str:
        """
        Returns the address of the Autonity contract.
        """
        addr = self._getContractAddress()
        assert isinstance(addr, str)
        return self.w3.to_checksum_address(addr)

    def get_core_state(self) -> Any:
        """
        Returns the core state of attached node.
        """
        core_state = self._getCoreState()
        print(f"core_state={core_state}")
        return core_state

    def get_committee_enodes(self) -> Sequence[str]:
        """
        Returns the set of enodes in the current consensus committee.
        """
        committee_enodes = self._getCommitteeEnodes()
        print(f"core_state={committee_enodes}")
        return committee_enodes



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Committee member model class
"""

from __future__ import annotations
from web3.types import ChecksumAddress
from typing import TypedDict, Tuple


class CommitteeMember(TypedDict):
    """
    Dictionary representing a member of the consensus committee.
    """

    address: ChecksumAddress
    voting_power: int


def committee_member_from_tuple(value: Tuple[ChecksumAddress, int]) -> CommitteeMember:
    """
    From Web3 tuple
    """
    assert len(value) == 2
    assert isinstance(value[0], str)
    assert isinstance(value[1], int)
    return CommitteeMember({"address": value[0], "voting_power": value[1]})



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Top-level Autonity module exposing primary types.
"""

# pylint: disable=unused-import
# flake8: noqa

from autonity.autonity import Autonity, AUTONITY_CONTRACT_VERSION
from autonity.config import Config
from autonity.committee_member import CommitteeMember
from autonity.validator import ValidatorDescriptor, Validator
from autonity.tendermint import Tendermint



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Model for an ERC20 token
"""

from autonity.abi_manager import ABIManager

from web3.contract.contract import Contract, ABI, ContractFunction
from web3.exceptions import BadFunctionCallOutput, ContractLogicError
from web3.types import Address, ChecksumAddress, ABIFunction
from web3 import Web3
from typing import Union, Optional

# pylint: disable=too-many-arguments


NAME_FUNCTION: ABIFunction = {
    "inputs": [],
    "name": "name",
    "outputs": [{"internalType": "string", "name": "", "type": "string"}],  # type: ignore
    "stateMutability": "pure",
    "type": "function",
}

SYMBOL_FUNCTION: ABIFunction = {
    "inputs": [],
    "name": "symbol",
    "outputs": [{"internalType": "string", "name": "", "type": "string"}],  # type: ignore
    "stateMutability": "pure",
    "type": "function",
}

DECIMALS_FUNCTION: ABIFunction = {
    "inputs": [],
    "name": "decimals",
    "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],  # type: ignore
    "stateMutability": "pure",
    "type": "function",
}


def add_missing_erc20_functions(abi: ABI) -> ABI:
    """
    The IERC20 ABI doesn't include these optional functions by
    default.  We add them in to all ERC20 contract ABIs, and check
    them at runtime.
    """
    function_names = [fn["name"] for fn in abi if fn["type"] == "function"]
    abi = list(abi)
    if "name" not in function_names:
        abi.append(NAME_FUNCTION)
    if "symbol" not in function_names:
        abi.append(SYMBOL_FUNCTION)
    if "decimals" not in function_names:
        abi.append(DECIMALS_FUNCTION)

    return abi


class ERC20:
    """
    Wrapper class for an ERC20 contract
    """

    contract: Contract

    def __init__(
        self,
        web3: Web3,
        address: Union[Address, ChecksumAddress],
        abi: Optional[ABI] = None,
    ):
        if not abi:
            abi = add_missing_erc20_functions(ABIManager.load_abi("IERC20"))
        self.contract = web3.eth.contract(address, abi=abi)

    def name(self) -> Optional[str]:
        """
        Returns the token name (or None if not available).
        """
        # TODO try-catch here in case the function is not present.
        name_function = getattr(self.contract.functions, "name", None)
        try:
            return name_function().call() if name_function else None
        except (BadFunctionCallOutput, ContractLogicError):
            return None

    def symbol(self) -> Optional[str]:
        """
        Returns the token symbol (or None if not available).
        """
        symbol_function = getattr(self.contract.functions, "symbol", None)
        try:
            return symbol_function().call() if symbol_function else None
        except (BadFunctionCallOutput, ContractLogicError):
            return None

    def decimals(self) -> int:
        """
        Returns the number of decimals used in the token (or None if not available).
        """
        decimals_function = getattr(self.contract.functions, "decimals", None)
        try:
            return decimals_function().call() if decimals_function else 0
        except (BadFunctionCallOutput, ContractLogicError):
            # https://ethereum.stackexchange.com/questions/100039/whats-the-default-erc20-decimals
            return 0

    def total_supply(self) -> int:
        """
        Total supply in "token units" (divide by 10^decimals for value in whole tokens).
        """
        return self.contract.functions.totalSupply().call()

    def balance_of(self, account: ChecksumAddress) -> int:
        """
        Returns the balance of a particular address in "token units"
        (divide by 10^decimals for value in whole tokens).
        """
        return self.contract.functions.balanceOf(account).call()

    def allowance(self, owner: ChecksumAddress, spender: ChecksumAddress) -> int:
        """
        Returns the quantity that `owner` has granted `spender` permission
        to spend.  Given in "token units" (divide by 10^decimals for
        value in whole tokens).
        """
        return self.contract.functions.allowance(owner, spender).call()

    def transfer(
        self,
        recipient: ChecksumAddress,
        amount: int,
    ) -> ContractFunction:
        """
        Create a transaction transferring `amount` in "token units"
        (units of 1/10^decimals) to `recipient`.
        """
        return self.contract.functions.transfer(recipient, amount)

    def approve(
        self,
        spender: ChecksumAddress,
        amount: int,
    ) -> ContractFunction:
        """
        Create a transaction granting `spender` permission to spend
        `amount` in "token units" (units of 1/10^decimals) held by
        `from_addr`.
        """
        return self.contract.functions.approve(spender, amount)

    def transfer_from(
        self,
        spender: ChecksumAddress,
        recipient: ChecksumAddress,
        amount: int,
    ) -> ContractFunction:
        """
        Create a transaction transferring `amount` in "token units" (units
        of 1/10^decimals) of the tokens held by `spender` to
        `recipient`.  `spender` must previously have granted
        `from_addr` permission to spend these tokens, via an `approve`
        transaction.
        """
        return self.contract.functions.transferFrom(spender, recipient, amount)

    # TODO: expose events?
    #   event Transfer(address indexed from, address indexed to, uint256 value);
    #   event Approval(address indexed owner, address indexed spender, uint256 value);



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Keyfile utility functions
"""

from eth_account import Account
from eth_keyfile import create_keyfile_json, decode_keyfile_json  # type: ignore
from web3 import Web3
from web3.types import ChecksumAddress
import json
from typing import Dict, NewType, Any, cast


EncryptedKeyData = NewType("EncryptedKeyData", Dict[str, Any])

PrivateKey = NewType("PrivateKey", bytes)


def load_keyfile(filename: str) -> EncryptedKeyData:
    """
    Load a keyfile.
    """
    with open(filename, "r", encoding="utf8") as keyfile_f:
        keyfile_json = json.load(keyfile_f)
        # TODO: validate the json
        return cast(EncryptedKeyData, keyfile_json)


def get_address_from_private_key(private_key: PrivateKey) -> ChecksumAddress:
    """
    Return the address associated with a given private key.
    """
    account = Account.from_key(private_key)  # pylint: disable=no-value-for-parameter
    return account.address


def create_keyfile_from_private_key(
    private_key: PrivateKey, password: str
) -> EncryptedKeyData:
    """
    Create a keyfile (with encrypted private key) from a private key.
    """
    assert len(private_key) == 32
    return create_keyfile_json(private_key, password.encode("utf8"))


def get_address_from_keyfile(encrypted_key: EncryptedKeyData) -> ChecksumAddress:
    """
    Given some keyfile data, extract the address.
    """
    return Web3.to_checksum_address(encrypted_key["address"])


def decrypt_keyfile(encrypted_key: EncryptedKeyData, password: str) -> PrivateKey:
    """
    Decrypt the private key from a keyfile.
    """
    private_key = decode_keyfile_json(encrypted_key, password.encode("utf8"))
    assert isinstance(private_key, bytes)
    return cast(PrivateKey, private_key)



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Model holding Validator information.
"""

from __future__ import annotations

from autonity.liquid_newton import LiquidNewton
from autonity.erc20 import ERC20

from enum import IntEnum
from web3 import Web3
from web3.types import ChecksumAddress, Wei
from web3.contract.contract import ContractFunction
from typing import TypedDict, NewType, Tuple

# pylint: disable=too-many-instance-attributes


ValidatorAddress = NewType("ValidatorAddress", ChecksumAddress)


class ValidatorState(IntEnum):
    """
    The status of a Validator
    """

    ACTIVE = 0
    PAUSED = 1


class ValidatorDescriptor(TypedDict):
    """
    Dictionary representing the description of a Validator
    """

    treasury: ChecksumAddress
    addr: ValidatorAddress
    enode: str
    commission_rate: int
    bonded_stake: int
    total_slashed: int
    liquid_contract: ChecksumAddress
    liquid_supply: int
    registration_block: int
    state: ValidatorState


def validator_descriptor_from_tuple(
    value: Tuple[
        ChecksumAddress,
        ValidatorAddress,
        str,
        int,
        int,
        int,
        ChecksumAddress,
        int,
        int,
        ValidatorState,
    ]
) -> ValidatorDescriptor:
    """
    Create an instance from the tuple returned by Web3 contract calls.
    """
    assert len(value) == 10
    assert isinstance(value[0], str)
    assert isinstance(value[1], str)
    assert isinstance(value[2], str)
    assert isinstance(value[3], int)
    assert isinstance(value[4], int)
    assert isinstance(value[5], int)
    assert isinstance(value[6], str)
    assert isinstance(value[7], int)
    assert isinstance(value[8], int)
    assert isinstance(value[9], int)

    return ValidatorDescriptor(
        {
            "treasury": value[0],
            "addr": value[1],
            "enode": value[2],
            "commission_rate": value[3],
            "bonded_stake": value[4],
            "total_slashed": value[5],
            "liquid_contract": value[6],
            "liquid_supply": value[7],
            "registration_block": value[8],
            "state": ValidatorState(value[9]),
        }
    )


class Validator:
    """
    Information held about a Validator.
    """

    treasury: ChecksumAddress
    address: ValidatorAddress
    enode: str
    commission_rate: int
    bonded_stake: int
    total_slashed: int
    lnew_contract: ERC20
    liquid_supply: int
    registration_block: int
    state: ValidatorState

    _liquid_contract: LiquidNewton
    """
    Internally this provides access to the extra functions.  Publicly,
    LNEW is exposed as a plain ERC20 token, via lnew_contract.
    """

    def __init__(self, w3: Web3, vdesc: ValidatorDescriptor):
        self.treasury = vdesc["treasury"]
        self.address = vdesc["addr"]
        self.enode = vdesc["enode"]
        self.commission_rate = vdesc["commission_rate"]
        self.bonded_stake = vdesc["bonded_stake"]
        self.total_slashed = vdesc["total_slashed"]
        self._liquid_contract = LiquidNewton(w3, vdesc["liquid_contract"])
        self.lnew_contract = self._liquid_contract
        self.liquid_supply = vdesc["liquid_supply"]
        self.registration_block = vdesc["registration_block"]
        self.state = vdesc["state"]

    def unclaimed_rewards(self, account: ChecksumAddress) -> Wei:
        """
        Query the rewards for this validator, claimable by `account`.
        """
        return self._liquid_contract.unclaimed_rewards(account)

    def claim_rewards(self) -> ContractFunction:
        """
        Create a ContractFunction to claim the rewards due to th call, from this validator.
        """
        return self._liquid_contract.claim_rewards()



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Utilities related to formatting token and Auton balances.
Constants related to known token precisions.
"""

from decimal import Decimal
from web3.types import Wei


AUTON_DECIMALS = 18

# TODO: Change this to 18 once the Autonity node supports that.
NEWTON_DECIMALS = 18


def format_quantity(units: int, decimals: int) -> str:
    """
    Given some quantity of atomic "token units" of a currency
    (e.g. Attoton for Auton), and the decimals used to represent the
    value of such a unit, return a string representation of the number
    of tokens.
    """
    adjusted = Decimal(units) * pow(Decimal(10), -decimals)
    return f"{adjusted:f}"


def format_auton_quantity(wei: Wei) -> str:
    """
    Format a quantity of Wei as a string representing a quantity of
    Auton, with decimal places.
    """
    return format_quantity(wei, AUTON_DECIMALS)


def format_newton_quantity(units: int) -> str:
    """
    Format a quantity of Newton units as a string representing a
    quantity of whole Newton, with decimal places.
    """
    return format_quantity(units, NEWTON_DECIMALS)



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
ABIManager class
"""

import autonity.abi

import importlib.resources  # type: ignore
import json
from web3.contract.contract import ABI
from typing import Dict


_CACHE: Dict[str, ABI] = {}


class ABIManager:
    """
    Access to bundled contract ABIs
    """

    @staticmethod
    def load_abi(contract_name: str) -> ABI:
        """
        Load (and cache) the ABI for a protocol contract.
        """
        if contract_name not in _CACHE:
            text = importlib.resources.read_text(autonity.abi, contract_name + ".abi")
            _CACHE[contract_name] = json.loads(text)

        return _CACHE[contract_name]

    @staticmethod
    def load_abi_file(file_name: str) -> ABI:
        """
        Load an ABI from a file.
        """
        with open(file_name, "r", encoding="utf8") as abi_f:
            return json.load(abi_f)



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Python module holding the Autonity Web3.py external module
"""

from __future__ import annotations

from autonity.config import Config, config_from_tuple
from autonity.validator import (
    validator_descriptor_from_tuple,
    ValidatorDescriptor,
    ValidatorAddress,
)
from autonity.committee_member import CommitteeMember, committee_member_from_tuple
from autonity.erc20 import ERC20
from autonity.abi_manager import ABIManager

from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.types import ChecksumAddress, Wei, HexBytes
from typing import Sequence, TypedDict, Tuple, cast

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-arguments

AUTONITY_CONTRACT_VERSION = 2
"""
The version of the Autonity contract which this library is aligned
with.
"""


# TODO: use the tendermint RPC call for this?
AUTONITY_CONTRACT_ADDRESS = "0xBd770416a3345F91E4B34576cb804a576fa48EB1"


class Staking(TypedDict):
    """
    Queued staking operations
    """

    delegator: ChecksumAddress
    delegatee: ChecksumAddress
    amount: int
    start_block: int


def staking_from_tuple(value: Tuple[str, str, int, int]) -> Staking:
    """
    Create Staking object from a Web3 tuple.
    """
    assert len(value) == 4
    assert isinstance(value[0], str)
    assert isinstance(value[1], str)
    assert isinstance(value[2], int)
    assert isinstance(value[3], int)
    return Staking(
        {
            "delegator": cast(ChecksumAddress, value[0]),
            "delegatee": cast(ChecksumAddress, value[1]),
            "amount": value[2],
            "start_block": value[3],
        }
    )


class Autonity(ERC20):
    """
    Web3 module representing Autonity-specific API.
    """

    def __init__(self, web3: Web3):
        super().__init__(
            web3,
            web3.to_checksum_address(AUTONITY_CONTRACT_ADDRESS),
            ABIManager.load_abi("Autonity"),
        )

        # TODO: What contract version checks can be performed here?

    @staticmethod
    def address() -> ChecksumAddress:
        """
        Return the deterministic address of the Autonity contract.
        """
        return Web3.to_checksum_address(AUTONITY_CONTRACT_ADDRESS)

    def commission_rate_precision(self) -> int:
        """
        See `COMMISSION_RATE_PRECISION` on the Autonity contract.
        """
        return self.contract.functions.COMMISSION_RATE_PRECISION().call()

    def config(self) -> Config:
        """
        See `config` on the Autonity contract.
        """
        return config_from_tuple(self.contract.functions.config().call())

    def epoch_id(self) -> int:
        """
        See `epochID` on the Autonity contract.
        """
        return self.contract.functions.epochID().call()

    def last_epoch_block(self) -> int:
        """
        See `lastEpochBlock` on the Autonity contract.
        """
        return self.contract.functions.lastEpochBlock().call()

    def epoch_total_bonded_stake(self) -> int:
        """
        See `epochTotalBondedStake` on the Autonity contract.
        """
        return self.contract.functions.epochTotalBondedStake().call()

    def total_redistributed(self) -> int:
        """
        See `totalRedistributed` on the Autonity contract.
        """
        return self.contract.functions.totalRedistributed().call()

    def epoch_reward(self) -> int:
        """
        See `epochReward` on the Autonity contract.
        """
        return self.contract.functions.epochReward().call()

    def tail_bonding_id(self) -> int:
        """
        See `tailBondingID` on the Autonity contract.
        """
        return self.contract.functions.tailBondingID().call()

    def head_bonding_id(self) -> int:
        """
        See `headBondingID` on the Autonity contract.
        """
        return self.contract.functions.headBondingID().call()

    def tail_unbonding_id(self) -> int:
        """
        See `tailUnbondingID` on the Autonity contract.
        """
        return self.contract.functions.tailUnbondingID().call()

    def head_unbonding_id(self) -> int:
        """
        See `headUnbondingID` on the Autonity contract.
        """
        return self.contract.functions.headUnbondingID().call()

    def deployer(self) -> ChecksumAddress:
        """
        See `deployer` on the Autonity contract.
        """
        return self.contract.functions.deployer().call()

    def get_last_epoch_block(self) -> int:
        """
        See `getLastEpochBlock` on the Autonity contract.
        """
        return self.contract.functions.getLastEpochBlock().call()

    def get_version(self) -> int:
        """
        See `getVersion` on the Autonity contract.
        """
        return self.contract.functions.getVersion().call()

    def get_committee(self) -> Sequence[CommitteeMember]:
        """
        See `getCommittee` on the Autonity contract.
        """
        cms = self.contract.functions.getCommittee().call()
        return [committee_member_from_tuple(cm) for cm in cms]

    def get_validators(self) -> Sequence[ValidatorAddress]:
        """
        See `getValidators` on the Autonity contract.
        """
        return self.contract.functions.getValidators().call()

    def get_validator(self, addr: ChecksumAddress) -> ValidatorDescriptor:
        """
        See `getValidator` on the Autonity contract.  To interact with
        validators, construct a `Validator` object from the returned
        ValidatorDescription.
        """
        assert isinstance(addr, str)
        value = self.contract.functions.getValidator(addr).call()
        return validator_descriptor_from_tuple(value)

    def get_max_committee_size(self) -> int:
        """
        See `getMaxCommitteeSize` on the Autonity contract.
        """
        return self.contract.functions.getMaxCommitteeSize().call()

    def get_committee_enodes(self) -> Sequence[str]:
        """
        See `getCommitteeEnodes` on the Autonity contract.
        """
        return self.contract.functions.getCommitteeEnodes().call()

    def get_minimum_base_fee(self) -> int:
        """
        See `getMinimumBaseFee` on the Autonity contract.
        """
        return self.contract.functions.getMinimumBaseFee().call()

    def get_operator(self) -> ChecksumAddress:
        """
        See `getOperator` on the Autonity contract.
        """
        return self.contract.functions.getOperator().call()

    def get_proposer(self, height: int, round_idx: int) -> ChecksumAddress:
        """
        See `getProposer` on the Autonity contract.
        """
        return self.contract.functions.getProposer(height, round_idx).call()

    def get_bonding_req(self, start_id: int, last_id: int) -> Sequence[Staking]:
        """
        See `getBondingReq` on the Autonity contract.
        """
        result = self.contract.functions.getBondingReq(start_id, last_id).call()
        assert isinstance(result, list)
        return [staking_from_tuple(staking) for staking in result]

    def get_unbonding_req(self, start_id: int, last_id: int) -> Sequence[Staking]:
        """
        See `getUnbondingReq` on the Autonity contract.
        """
        result = self.contract.functions.getUnbondingReq(start_id, last_id).call()
        assert isinstance(result, list)
        return [staking_from_tuple(staking) for staking in result]

    def bond(
        self,
        validator_addr: ValidatorAddress,
        amount: int,
    ) -> ContractFunction:
        """
        Create a TxParams calling the `bond` method.  See `bond` on the
        Autonity contract.
        """
        return self.contract.functions.bond(validator_addr, amount)

    def unbond(
        self,
        validator_addr: ValidatorAddress,
        amount: int,
    ) -> ContractFunction:
        """
        Create a TxParams calling the `unbond` method.  See `unbond` on
        the Autonity contract.
        """
        return self.contract.functions.unbond(validator_addr, amount)

    def register_validator(self, enode: str, proof: HexBytes) -> ContractFunction:
        """
        Create a TxParams calling the `registerValidator` method.  See
        `registerValidator` on the Autonity contract.
        """
        return self.contract.functions.registerValidator(enode, proof)

    def pause_validator(
        self,
        validator_addr: ValidatorAddress,
    ) -> ContractFunction:
        """
        Create a TxParams calling the `pauseValidator` method.  See
        `pauseValidator` on the Autonity contract.
        """
        return self.contract.functions.pauseValidator(validator_addr)

    def activate_validator(
        self,
        validator_addr: ValidatorAddress,
    ) -> ContractFunction:
        """
        Create a TxParams calling the `activateValidator` method.  See
        `activateValidator` on the Autonity contract.
        """
        return self.contract.functions.activateValidator(validator_addr)

    # Functions below here are operatorOnly

    def set_minimum_base_fee(self, base_fee: Wei) -> ContractFunction:
        """
        Set the minimum gas price. Restricted to the operator account.
        See `setMinimumBaseFee` on the Autonity contract.
        """
        return self.contract.functions.setMinimumBaseFee(base_fee)

    def set_committee_size(self, committee_size: int) -> ContractFunction:
        """
        Set the maximum size of the consensus committee. Restricted to the
        Operator account.  See `setCommitteeSize` on Autonity contract.
        """
        return self.contract.functions.setCommitteeSize(committee_size)

    def set_unbonding_period(self, period: int) -> ContractFunction:
        """
        Set the unbonding period. Restricted to the Operator account.  See
        `setUnbondingPeriod` on Autonity contract.
        """
        return self.contract.functions.setUnbondingPeriod(period)

    def set_epoch_period(self, period: int) -> ContractFunction:
        """
        Set the epoch period. Restricted to the Operator account.  See
        `setEpochPeriod` on Autonity contract.
        """
        return self.contract.functions.setEpochPeriod(period)

    def set_operator_account(self, address: ChecksumAddress) -> ContractFunction:
        """
        Set the Operator account. Restricted to the Operator account.  See
        `setOperatorAccount` on Autonity contract.
        """
        return self.contract.functions.setOperatorAccount(address)

    # def set_block_period(period: int) -> ContractFunction:
    #     """
    #     Currently not supported.  Set the block period. Restricted to the
    #     Operator account.
    #     """
    #     return self.contract.setBlockPeriod(period)

    def set_treasury_account(self, address: ChecksumAddress) -> ContractFunction:
        """
        Set the global treasury account. Restricted to the Operator
        account.  See `setTreasuryAccount` on Autonity contract.
        """
        return self.contract.functions.setTreasuryAccount(address)

    def set_treasury_fee(self, treasury_fee: int) -> ContractFunction:
        """
        Set the treasury fee. Restricted to the Operator account.  See
        `setTreasuryFee` on Autonity contract.
        """
        return self.contract.functions.setTreasuryFee(treasury_fee)

    def mint(self, address: ChecksumAddress, amount: int) -> ContractFunction:
        """
        Mint new stake token (NTN) and add it to the recipient
        balance. Restricted to the Operator account.  See `mint` on
        Autonity contract.
        """
        return self.contract.functions.mint(address, amount)

    def burn(self, address: ChecksumAddress, amount: int) -> ContractFunction:
        """
        Burn the specified amount of NTN stake token from an
        account. Restricted to the Operator account.  This won't burn
        associated Liquid tokens.  See `burn` on Autonity contract.
        """
        return self.contract.functions.burn(address, amount)

    def change_commission_rate(
        self,
        validator: ValidatorAddress,
        rate: int,
    ) -> ContractFunction:
        """
        Create a TxParams calling the `changeCommissionRate` method.  See
        `changeCommissionRate` on the Autonity contract.
        """
        return self.contract.functions.changeCommissionRate(validator, rate)

    # TODO: events

    # event MintedStake(address addr, uint256 amount);
    # event BurnedStake(address addr, uint256 amount);
    # event CommissionRateChange(address validator, uint256 rate);
    # event RegisteredValidator(
    #   address treasury, address addr, string enode, address liquidContract);
    # event PausedValidator(address treasury, address addr, uint256 effectiveBlock);
    # event Rewarded(address addr, uint256 amount);



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Web3 utility functions
"""

from autonity.tendermint import Tendermint

import re
from web3 import Web3
from web3.providers import BaseProvider
from web3.module import Module
from typing import Dict, Sequence, Union, Optional, Type, Any, cast


class Web3WithAutonity(Web3):
    """
    Web3.py class with Autonity-related modules. Used to inform the
    static type checker of the existence and types of the modules.
    """

    tendermint: Tendermint


def web3_provider_for_endpoint(endpoint: str) -> BaseProvider:
    """
    Given an rpc endpoint, return an appropriate provider (https, ws,
    or IPC). If identifier isn't a valid format of one of these three
    types, throws an exception.
    """
    regex_http = re.compile(r"^(?:http)s?://")
    if re.match(regex_http, endpoint) is not None:
        return Web3.HTTPProvider(endpoint)

    regex_ws = re.compile(r"^(?:ws)s?://")
    if re.match(regex_ws, endpoint) is not None:
        return Web3.WebsocketProvider(endpoint)

    regex_ipc = re.compile("([^ !$`&*()+]|(\\[ !$`&*()+]))+\\.ipc")
    if re.match(regex_ipc, endpoint) is not None:
        return Web3.IPCProvider(endpoint)

    raise ValueError(f"cannot determine provider for: {endpoint}")


def create_web3(
    provider: Optional[BaseProvider] = None,
    external_modules: Optional[Dict[str, Union[Type[Module], Sequence[Any]]]] = None,
    ignore_chain_id: bool = True,
    **kwArgs: Any,
) -> Web3WithAutonity:
    """
    Convenience function to create a Web3 instance with the Autonity
    external modules attached.  Returns a Web3WithAut type, for static
    type checking.
    """

    external_modules = external_modules or {}
    external_modules["tendermint"] = Tendermint
    w3 = Web3(
        provider,
        external_modules={
            "tendermint": Tendermint,
        },
        **kwArgs,
    )

    # Check the chain ID and ensure it conforms to the Autonity
    # standard.
    if not ignore_chain_id:
        chain_id = w3.eth.chain_id
        _ = parse_autonity_chain_id(chain_id)

    return cast(Web3WithAutonity, w3)


def create_web3_for_endpoint(
    endpoint: str, ignore_chain_id: bool = True, **kwArgs: Any
) -> Web3WithAutonity:
    """
    Convenience function to create a Web3 object for a specific
    endpoint URL.
    """
    return create_web3(
        web3_provider_for_endpoint(endpoint, **kwArgs), ignore_chain_id=ignore_chain_id
    )


def parse_autonity_chain_id(chain_id: int) -> str:
    """
    Parse a chain ID according to the Autonity scheme and return a
    human-readable version.  Raise an exception if the chain ID is not
    recognised.
    """

    # As a decimal: '65xxyyyy' where 'xx' = network type and 'yyyy' =
    # identifier.

    digits = str(chain_id)
    if len(digits) != 8 or digits[:2] != "65":
        raise ValueError("chain ID does not match the Autonity scheme")

    net_type = digits[2:4]
    net_id = digits[4:]

    return f"Autonity (type: {net_type}, id: {net_id})"



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Models for Autonity contract configuration
"""

from __future__ import annotations

from web3.types import ChecksumAddress
from typing import TypedDict, Tuple


class Config(TypedDict):
    """
    Autonity configuration.
    """

    operator_account: ChecksumAddress
    treasury_account: ChecksumAddress
    treasury_fee: int
    min_basefee: int
    delegation_rate: int
    epoch_period: int
    unbonding_period: int
    committee_size: int
    contract_version: int
    block_period: int


def config_from_tuple(
    value: Tuple[
        ChecksumAddress, ChecksumAddress, int, int, int, int, int, int, int, int
    ]
) -> Config:
    """
    Create from a Web3 tuple.
    """
    assert isinstance(value[0], str)
    assert isinstance(value[1], str)
    assert isinstance(value[2], int)
    assert isinstance(value[3], int)
    assert isinstance(value[4], int)
    assert isinstance(value[5], int)
    assert isinstance(value[6], int)
    assert isinstance(value[7], int)
    assert isinstance(value[8], int)
    assert isinstance(value[9], int)
    return Config(
        {
            "operator_account": value[0],
            "treasury_account": value[1],
            "treasury_fee": value[2],
            "min_basefee": value[3],
            "delegation_rate": value[4],
            "epoch_period": value[5],
            "unbonding_period": value[6],
            "committee_size": value[7],
            "contract_version": value[8],
            "block_period": value[9],
        }
    )



# Copyright (C) 2015-2022 Clearmatics Technologies Ltd - All Rights Reserved.

"""
Transaction utility functions
"""

from autonity.utils.keyfile import (
    EncryptedKeyData,
    PrivateKey,
    decrypt_keyfile,
)

from eth_account.account import Account, SignedTransaction  # type: ignore
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.types import ChecksumAddress, TxParams, TxReceipt, Wei, Nonce, HexBytes
from web3._utils.transactions import fill_transaction_defaults
from typing import Callable, Optional

# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches


def create_transaction(
    from_addr: Optional[ChecksumAddress] = None,
    to_addr: Optional[ChecksumAddress] = None,
    value: Optional[Wei] = None,
    data: Optional[HexBytes] = None,
    gas: Optional[Wei] = None,
    gas_price: Optional[Wei] = None,
    max_fee_per_gas: Optional[Wei] = None,
    max_priority_fee_per_gas: Optional[Wei] = None,
    nonce: Optional[Nonce] = None,
    chain_id: Optional[int] = None,
) -> TxParams:
    """
    Optionally set some fields on a TxParams object.  This can then be
    passed to contract call methods, and then any remaining fields set
    with finalize_transaction.
    """
    tx: TxParams = {}

    if from_addr:
        tx["from"] = from_addr

    if to_addr:
        tx["to"] = to_addr

    if value:
        tx["value"] = value

    if data:
        tx["data"] = data

    if nonce:
        tx["nonce"] = nonce

    if gas:
        tx["gas"] = gas

    if chain_id:
        tx["chainId"] = chain_id

    # Require either gas_price OR max_fee_per_gas, etc

    if gas_price:
        if max_fee_per_gas or max_priority_fee_per_gas:
            raise ValueError("gas price cannot be used with other fee parameters")
        tx["gasPrice"] = gas_price
    else:
        if max_fee_per_gas:
            tx["maxFeePerGas"] = max_fee_per_gas

            if max_priority_fee_per_gas:
                tx["maxPriorityFeePerGas"] = max_priority_fee_per_gas
            else:
                tx["maxPriorityFeePerGas"] = tx["maxFeePerGas"]

    return tx


def create_contract_function_transaction(
    function: ContractFunction,
    from_addr: ChecksumAddress,
    value: Optional[Wei] = None,
    gas: Optional[Wei] = None,
    gas_price: Optional[Wei] = None,
    max_fee_per_gas: Optional[Wei] = None,
    max_priority_fee_per_gas: Optional[Wei] = None,
    nonce: Optional[Nonce] = None,
    chain_id: Optional[int] = None,
) -> TxParams:
    """
    Given a contract call and other parameters, create an unsigned
    transaction (Web3 TxParams object).  Any fields not passed in will
    be filled out by querying the attached node, if available.
    """
    tx = create_transaction(
        from_addr=from_addr,
        value=value,
        gas=gas,
        gas_price=gas_price,
        max_fee_per_gas=max_fee_per_gas,
        max_priority_fee_per_gas=max_priority_fee_per_gas,
        nonce=nonce,
        chain_id=chain_id,
    )

    return function.build_transaction(tx)


def finalize_transaction(
    create_w3: Callable[[], Web3],
    tx: TxParams,
    from_addr: Optional[ChecksumAddress],
) -> TxParams:
    """
    Fill in any values not already set.  If necessary, a Web3 object
    will be created via the create_w3 callback.
    """

    w3: Optional[Web3] = None

    def get_web3() -> Web3:
        nonlocal w3
        if not w3:
            w3 = create_w3()
        return w3

    if "gas" not in tx:
        w3 = get_web3()
        tx["gas"] = w3.eth.estimate_gas(tx)

    if "nonce" not in tx:
        if not from_addr:
            raise ValueError("neither nonce or from-address given")
        w3 = get_web3()
        tx["nonce"] = w3.eth.get_transaction_count(from_addr)

    if "chainId" not in tx:
        w3 = get_web3()
        tx["chainId"] = w3.eth.chain_id

    if "gasPrice" not in tx and "maxFeePerGas" not in tx:
        tx = fill_transaction_defaults(get_web3(), tx)

    # The code above to fill defaults introduces an (unserializable)
    # empty bytes() in the "data" field.

    if "data" in tx:
        data = tx["data"]
        if len(data) == 0:
            del tx["data"]
        elif isinstance(data, bytes):
            tx["data"] = Web3.to_hex(data)

    return tx


def sign_tx_with_private_key(
    tx: TxParams, private_key: PrivateKey
) -> SignedTransaction:
    """
    For cases where the private key has already been decrypted.  (In
    general, for signing single transactions `sign_tx` is
    recommended.)
    """
    return Account.sign_transaction(  # pylint: disable=no-value-for-parameter
        tx, private_key
    )


def sign_tx(
    tx: TxParams, key_data: EncryptedKeyData, key_passphrase: str
) -> SignedTransaction:
    """
    Sign a transaction using the data from an encrypted keyfile and
    password.
    """
    private_key = decrypt_keyfile(key_data, key_passphrase)
    return sign_tx_with_private_key(tx, private_key)


def send_tx(w3: Web3, tx_signed: SignedTransaction) -> HexBytes:
    """
    Send raw signed tx bytes provided by 'tx_raw' to an RPC server
    to be validated and included in the blockchain.
    """
    raw_tx = tx_signed.rawTransaction
    tx_hash = w3.eth.send_raw_transaction(raw_tx)
    return tx_hash


def wait_for_tx(w3: Web3, tx_hash: HexBytes, timeout: Optional[float]) -> TxReceipt:
    """
    Wait for a specific transaction.  Returns the receipts.

    This simply wraps Web3.eth.wait_for_transaction_receipt, but uses
    stricter types (accepts only a HexBytes object).
    """
    if timeout is None:
        return w3.eth.wait_for_transaction_receipt(tx_hash)

    return w3.eth.wait_for_transaction_receipt(tx_hash, timeout=timeout)
